# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
from cryptography import x509
from cryptography.x509.oid import NameOID
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes, serialization
from cryptography.hazmat.primitives.asymmetric import rsa
import datetime
import socket
import ipaddress


def get_my_ip():
    """Returns the IP address of the machine running the FastAPI server."""
    try:
        # Connect to an external server (Google's DNS) to determine our outward-facing IP
        s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        s.connect(("8.8.8.8", 80))
        ip = s.getsockname()[0]
        s.close()
    except Exception:
        ip = "127.0.0.1"  # Fallback to localhost if detection fails
    return ip


def generate_ssl_certificate(ip: str | None = None, domain: str | None = None,
                             key_file_name="key.pem", cert_file_name="cert.pem"):
    """
    Generates a new SSL certificate and saves it to a file
    :param ip: IP address of the server (optional)
    :param domain: Domain name of the server (optional)
    :param key_file_name: name of the key file
    :param cert_file_name: name of the certificate file
    """

    subjects_list = [x509.DNSName(u"localhost")]

    if ip is not None:
        subjects_list.append(x509.IPAddress(ipaddress.ip_address(ip)))

    if domain is not None:
        subjects_list.append(x509.DNSName(domain))

    # Generate a private key
    private_key = rsa.generate_private_key(
        public_exponent=65537,
        key_size=2048,
        backend=default_backend()
    )

    # Write the private key to a file
    with open(key_file_name, "wb") as key_file:
        key_file.write(private_key.private_bytes(
            encoding=serialization.Encoding.PEM,
            format=serialization.PrivateFormat.TraditionalOpenSSL,
            encryption_algorithm=serialization.NoEncryption()
        ))

    # Define certificate subject and issuer details
    subject = issuer = x509.Name([
        x509.NameAttribute(NameOID.COUNTRY_NAME, u"ES"),
        x509.NameAttribute(NameOID.STATE_OR_PROVINCE_NAME, u"Barcelona"),
        x509.NameAttribute(NameOID.LOCALITY_NAME, u"Barcelona"),
        x509.NameAttribute(NameOID.ORGANIZATION_NAME, u"eRoots Analytics"),
        x509.NameAttribute(NameOID.COMMON_NAME, ip),
    ])

    # Build the certificate
    cert = x509.CertificateBuilder().subject_name(
        subject
    ).issuer_name(
        issuer
    ).public_key(
        private_key.public_key()
    ).serial_number(
        x509.random_serial_number()
    ).not_valid_before(
        datetime.datetime.now(datetime.UTC)
    ).not_valid_after(
        # Certificate is valid for 365 days
        datetime.datetime.now(datetime.UTC) + datetime.timedelta(days=365)
    ).add_extension(
        x509.SubjectAlternativeName(subjects_list),
        critical=False,
    ).sign(private_key, hashes.SHA256(), default_backend())

    # Write the certificate to a file
    with open(cert_file_name, "wb") as cert_file:
        cert_file.write(cert.public_bytes(serialization.Encoding.PEM))

    print(f"SSL certificate and key have been generated: cert={cert_file_name}, key={key_file_name}")


if __name__ == "__main__":
    generate_ssl_certificate(ip="127.0.0.1", key_file_name="key.pem")



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0

import socket
import requests
from fastapi import APIRouter

from VeraGridServer.settings import settings


def get_local_ip():
    try:
        return socket.gethostbyname(socket.gethostname())  # Gets internal IP
    except Exception:
        return "127.0.0.1"  # Fallback to localhost


def register_with_master(host: str, port: int, this_port: int, username: str, password: str):
    """
    register this service with a master service
    :param host:
    :param port:
    :param this_port:
    :param username:
    :param password:
    :return:
    """

    url = f"{host}:{port}/register_child_server"

    worker_data = {
        "ip": get_local_ip(),  # Detect IP dynamically
        "port": this_port,  # Set worker's port
        "username": username,
        "password": password,
    }

    response = requests.post(url,
                             json=worker_data,
                             verify=False,
                             timeout=5)
    response.raise_for_status()
    return response.json()  # Return JSON response


def lifespan(app: APIRouter):
    """
    Function that is called once the app starts.
    :param app:
    :return:
    """
    if not settings.am_i_master:
        print("I am a child service")
        res = register_with_master(host=settings.master_host,
                                   port=settings.master_port,
                                   this_port=settings.this_port,
                                   username=settings.this_username,
                                   password=settings.this_password)
        print("registering response:", res)
    else:
        print("I am the master service")

    yield  # The application runs here


router = APIRouter(lifespan=lifespan)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
import os
import sys
import uvicorn
import argparse

PACKAGE_PARENT = '..'
SCRIPT_DIR = os.path.dirname(os.path.realpath(os.path.join(os.getcwd(), os.path.expanduser(__file__))))
sys.path.append(os.path.normpath(os.path.join(SCRIPT_DIR, PACKAGE_PARENT)))

from VeraGridServer.main import app
from VeraGridServer.settings import settings
from VeraGridServer.generate_ssl_key import generate_ssl_certificate, get_my_ip
from VeraGridServer.__version__ import __VeraGridServer_VERSION__


def start_server(key_file_name: str = "key.pem", cert_file_name: str = "cert.pem",
                 port: int = 8000, domain="localhost",
                 master_host: str = "", master_port: int = 0,
                 username: str = "", password: str = "", is_master: bool = True,
                 secure: bool = True):
    """
    Start server function
    :param key_file_name: name of the key file that the server generates
    :param cert_file_name: name of the certificate file that the server generates
    :param port: Port to serve (8000 usually)
    :param domain: Domain to serve (i.e. localhost)
    :param master_host: IP address to register the server to (if this runs in child mode)
    :param master_port: Port to register the server to (if this runs in child mode)
    :param username: Username to authenticate with
    :param password: Password to authenticate with
    :param is_master: Whether the server is master or not
    :param secure: Whether the server is secure or not (if it looks for the certificates or not)
    """

    # find out my IP
    host = get_my_ip()

    print(f"""
┓┏      ┏┓  • ┓┏┓({__VeraGridServer_VERSION__} Alpha) 
┃┃┏┓┏┓┏┓┃┓┏┓┓┏┫┗┓┏┓┏┓┓┏┏┓┏┓
┗┛┗ ┛ ┗┻┗┛┛ ┗┗┻┗┛┗ ┛ ┗┛┗ ┛ 
{host}:{port}  
    """)

    if secure:
        generate_ssl_certificate(
            ip=host,
            domain=domain,
            key_file_name=key_file_name,
            cert_file_name=cert_file_name
        )

    # extra attributed on launch
    settings.am_i_master = is_master
    settings.master_host = master_host
    settings.master_port = master_port
    settings.this_host = host
    settings.this_port = port
    settings.this_username = username
    settings.this_password = password

    if secure:
        uvicorn.run(app,
                    host=host, port=port, ssl_keyfile=key_file_name, ssl_certfile=cert_file_name)
    else:
        uvicorn.run(app,
                    host=host, port=port)

def str2bool(value):
    if isinstance(value, bool):
        return value
    if value.lower() in {'True', 'true', 't', 'yes', '1'}:
        return True
    elif value.lower() in {'False', 'false', 'f', 'no', '0'}:
        return False
    else:
        raise argparse.ArgumentTypeError(f'Invalid boolean value: {value}')

if __name__ == "__main__":
    # Initialize parser
    parser = argparse.ArgumentParser(description="Start a secure VeraGrid server")

    # Add arguments
    parser.add_argument("--key_fname", type=str, default="key.pem",
                        help="Path to the private key file that the server generates")
    parser.add_argument("--cert_fname", type=str, default="cert.pem",
                        help="Path to the certificate file that the server generates")
    parser.add_argument("--host", type=str, default="0.0.0.0", help="Host IP address")
    parser.add_argument("--port", type=int, default=8000, help="Port to run the server on")

    parser.add_argument("--secure", type=str2bool,
                        choices=[True, False], default=True, help="Use https?")

    parser.add_argument("--master", type=str2bool,
                        choices=[True, False], default=True, help="Use https?")

    parser.add_argument("--master_host", type=str, default="0.0.0.0", help="URL of the master instance")
    parser.add_argument("--master_port", type=int, default=80, help="Port of the master instance")
    parser.add_argument("--user", type=str, default="", help="username")
    parser.add_argument("--pwd", type=str, default="", help="Password")

    # Parse arguments
    args = parser.parse_args()

    print("Arguments:")
    print('\n'.join(f'{k}: {v}' for k, v in vars(args).items()))

    # Call the start_server function with the parsed arguments
    start_server(key_file_name=args.key_fname,
                 cert_file_name=args.cert_fname,
                 port=args.port,
                 domain=args.host,
                 master_host=args.master_host,
                 master_port=args.master_port,
                 secure=args.secure,
                 is_master=args.master,
                 username=args.user,
                 password=args.pwd)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0

class ExtraSettings:
    def __init__(self):
        self.am_i_master: bool = True
        self.master_host: str = ""
        self.master_port = 0
        self.this_port = 0
        self.this_username = ""
        self.this_password = ""


settings = ExtraSettings()


# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
import os
import asyncio
import websockets
from typing import Callable


async def send_large_file(file_path: str,
                          websocket: websockets.WebSocketClientProtocol,
                          progress_func: Callable[[float], None] = None) -> None:
    """
    Send a large file via web-socket
    :param file_path: local file path
    :param websocket: web socket instance
    :param progress_func: Progress function
    :return: None
    """
    # Get the size of the file
    file_size = os.path.getsize(file_path)
    bytes_sent = 0

    # Open the file in binary mode
    with open(file_path, "rb") as file:
        # Read the file in chunks
        while True:
            # Read 4KB of data
            data = file.read(4096)
            if not data:
                break  # End of file

            # Send the chunk over the WebSocket
            await websocket.send(data)

            # Update bytes sent
            bytes_sent += len(data)

            # Calculate progress
            progress = (bytes_sent / file_size) * 100

            if progress_func is not None:
                progress_func(progress)


async def connect_and_send(file_path: str, url: str = 'ws://localhost:8000/ws'):
    """
    Connect and send data to web-socket
    """
    async with websockets.connect(url) as websocket:
        await send_large_file(file_path, websocket)


if __name__ == '__main__':
    asyncio.get_event_loop().run_until_complete(connect_and_send(file_path="my_file.dat", url="ws://localhost:8000"))
